// src/client.ts
import { EJSON } from "bson";

// src/errors.ts
var MongoDataAPIError = class extends Error {
  constructor(message, code) {
    super(message);
    this.code = code;
  }
};

// src/client.ts
import { EJSON as EJSON2, ObjectId } from "bson";
var removeEmptyKeys = (object) => {
  return Object.fromEntries(Object.entries(object).filter(([_, v]) => v));
};
var normalizeRequestOptions = (options) => {
  const { endpoint, dataSource, auth, fetch: customFetch } = options ?? {};
  const ftch = customFetch ?? globalThis.fetch;
  const connection = {
    dataSource,
    endpoint: endpoint instanceof URL ? endpoint.toString() : endpoint,
    headers: {},
    fetch: ftch
  };
  if (auth) {
    if ("apiKey" in auth) {
      connection.headers.apiKey = auth.apiKey;
    } else if ("jwtTokenString" in auth) {
      connection.headers.jwtTokenString = auth.jwtTokenString;
    } else if ("email" in auth && "password" in auth) {
      connection.headers.email = auth.email;
      connection.headers.password = auth.password;
    } else if ("bearerToken" in auth) {
      connection.headers.Authorization = `Bearer ${auth.bearerToken}`;
    } else {
      throw new Error("Invalid auth options");
    }
  }
  return connection;
};
var MongoClient = class {
  constructor(options) {
    this.connection = normalizeRequestOptions(options);
  }
  /** Select a database from within the data source */
  db(name) {
    return new Database(name, this.connection);
  }
};
var Database = class {
  constructor(name, connection) {
    this.connection = {
      ...connection,
      database: name
    };
  }
  /**
   * Get a Collection from the Database
   * @param name The collection name
   * @typeParam `T` The type of document stored in the collection
   * @returns A Collection object of type `T`
   */
  collection(name) {
    return new Collection(name, this.connection);
  }
};
var Collection = class {
  constructor(name, database) {
    this.connection = {
      ...database,
      collection: name
    };
  }
  async findOne(...args) {
    const [operationName, filter, options] = typeof args[0] === "string" ? args : ["default", ...args];
    const { data, error } = await this.callApi(
      "findOne",
      {
        filter,
        projection: options?.projection,
        sort: options?.sort
      },
      { operationName }
    );
    if (data) {
      return { data: data.document ?? null, error };
    }
    return { data, error };
  }
  async find(...args) {
    const [operationName, filter, options] = typeof args[0] === "string" ? args : ["default", ...args];
    const { data, error } = await this.callApi(
      "find",
      {
        filter,
        projection: options?.projection,
        sort: options?.sort,
        limit: options?.limit,
        skip: options?.skip
      },
      {
        operationName
      }
    );
    if (data) {
      return { data: data.documents, error };
    }
    return { data, error };
  }
  async insertOne(...args) {
    const [operationName, document] = typeof args[0] === "string" ? args : ["default", ...args];
    return this.callApi(
      "insertOne",
      {
        document
      },
      {
        operationName
      }
    );
  }
  async insertMany(...args) {
    const [operationName, documents] = typeof args[0] === "string" ? args : ["default", ...args];
    return this.callApi("insertMany", { documents }, { operationName });
  }
  async updateOne(...args) {
    const [operationName, filter, update, options] = typeof args[0] === "string" ? args : ["default", ...args];
    return this.callApi(
      "updateOne",
      {
        filter,
        update,
        upsert: options?.upsert
      },
      {
        operationName
      }
    );
  }
  async updateMany(...args) {
    const [operationName, filter, update, options] = typeof args[0] === "string" ? args : ["default", ...args];
    return this.callApi(
      "updateMany",
      {
        filter,
        update,
        upsert: options?.upsert
      },
      {
        operationName
      }
    );
  }
  async replaceOne(...args) {
    const [operationName, filter, replacement, options] = typeof args[0] === "string" ? args : ["default", ...args];
    return this.callApi(
      "replaceOne",
      {
        filter,
        replacement,
        upsert: options?.upsert
      },
      {
        operationName
      }
    );
  }
  async deleteOne(...args) {
    const [operationName, filter] = typeof args[0] === "string" ? args : ["default", ...args];
    return this.callApi(
      "deleteOne",
      {
        filter
      },
      {
        operationName
      }
    );
  }
  async deleteMany(...args) {
    const [operationName, filter] = typeof args[0] === "string" ? args : ["default", ...args];
    return this.callApi(
      "deleteMany",
      { filter },
      {
        operationName
      }
    );
  }
  async aggregate(...args) {
    const [operationName, pipeline] = typeof args[0] === "string" ? args : ["default", ...args];
    const response = await this.callApi(
      "aggregate",
      {
        pipeline
      },
      {
        operationName
      }
    );
    return response.data ? { data: response.data.documents } : {
      error: response.error ?? new MongoDataAPIError("Unknown error", -1)
    };
  }
  /**
   * Call a raw Data API resource
   * @param method The Data API resource to call
   * @param body The JSON body to include, merged with the collection, dataSource, and database names
   * @returns The response from the Data API, or a DataAPIError if the request failed
   */
  async callApi(method, body, requestOptions) {
    const {
      endpoint,
      dataSource,
      headers: h,
      collection,
      database,
      fetch: ftch
    } = this.connection;
    const headers = new Headers(h);
    headers.set("content-type", "application/ejson");
    if (requestOptions.operationName) {
      headers.set("x-realm-op-name", requestOptions.operationName);
    }
    const url = `${endpoint}/action/${method}`;
    const response = await ftch(url, {
      method: "POST",
      headers,
      body: EJSON.stringify({
        collection,
        database,
        dataSource,
        ...removeEmptyKeys(body)
      })
    });
    if (response.status < 200 || response.status >= 400) {
      let errorText;
      try {
        errorText = (await response.clone().json())?.error;
      } catch {
        try {
          errorText = await response.clone().text();
        } catch {
          errorText = void 0;
        }
      }
      const fallbackMessage = {
        400: "Bad Request",
        401: "Unauthorized",
        404: "Not Found",
        500: "Internal Server Error"
      };
      return {
        error: new MongoDataAPIError(
          errorText ?? response.statusText ?? fallbackMessage?.[response.status] ?? "Data API Error",
          response.status
        )
      };
    }
    const text = await response.text();
    return { data: EJSON.parse(text) };
  }
};
export {
  Collection,
  Database,
  EJSON2 as EJSON,
  MongoClient,
  MongoDataAPIError,
  ObjectId
};
//# sourceMappingURL=index.js.map